import logging
from typing import TYPE_CHECKING, Final

from uv_upx.services.collect_dependencies.collect_groups_from_py_project import collect_from_py_project
from uv_upx.services.dependency_up.constants.operators import VERSION_OPERATORS_I_PUT_IF_DIFFERENT
from uv_upx.services.dependency_up.parse_dependency import parse_dependency
from uv_upx.services.get_all_pyprojects import get_all_pyprojects_by_project_root_path

if TYPE_CHECKING:
    import pathlib
    from collections.abc import Iterable

    from uv_upx.services.dependency_up.models.dependencies_list import TomlBasedDependenciesList
    from uv_upx.services.dependency_up.models.dependency_parsed import DependencyParsed

TAB_CHARS: Final[str] = "  "


def filter_dependencies(
    dependencies: TomlBasedDependenciesList,
    *,
    only_special_cases: bool = False,
    #
    preserve_original_package_names: bool = False,
) -> Iterable[DependencyParsed]:
    for dep in dependencies:  # pyright: ignore[reportUnknownVariableType]
        if not isinstance(dep, str):
            continue

        dependency_parsed = parse_dependency(
            dep,
            preserve_original_package_names=preserve_original_package_names,
        )
        if only_special_cases and (
            (dependency_parsed.marker is not None)  # Have marker. It is an unusual case.
            or (len(dependency_parsed.version_constraints) > 1)  # Have more than one version constraint. It's complex.
            or (len(dependency_parsed.version_constraints) == 0)  # No version constraints. It's strange.
            or (
                any(
                    vc.operator not in VERSION_OPERATORS_I_PUT_IF_DIFFERENT
                    for vc in dependency_parsed.version_constraints
                )
            )
        ):
            yield dependency_parsed

        continue


def collect_top_level_dependencies(
    *,
    project_root_path: pathlib.Path,
    only_special_cases: bool = False,
    preserve_original_package_names: bool = False,
) -> None:
    logger = logging.getLogger(__name__)

    logger.info("Collecting top level dependencies...")

    py_projects = get_all_pyprojects_by_project_root_path(project_root_path)

    for py_project in py_projects.items:
        is_py_project_info_shown = False

        # Iterates dependencies; prints project, group, and dependency info
        for group in collect_from_py_project(py_project.data):
            is_group_title_shown = False

            for dep in filter_dependencies(
                group.dependencies,
                only_special_cases=only_special_cases,
                preserve_original_package_names=preserve_original_package_names,
            ):
                if not is_py_project_info_shown:
                    print(py_project.path.as_uri())
                    is_py_project_info_shown = True

                if not is_group_title_shown:
                    group_title = str(group.section)
                    if group.group_name:
                        group_title += f"[{group.group_name}]"

                    print(f"{TAB_CHARS}{group_title}:")
                    is_group_title_shown = True

                print(f"{TAB_CHARS * 2}{dep.get_full_spec()}")

            if is_group_title_shown:
                print("-" * 10)

        if is_py_project_info_shown:
            print("=" * 20)

    logger.info("Done.")



import copy
import logging
from html import escape
from typing import TYPE_CHECKING

from prompt_toolkit import HTML, choice, print_formatted_text

from uv_upx.services.dependency_up.update_dependency import update_dependency_v2
from uv_upx.services.toml import toml_save
from uv_upx.services.upgrade_profile import UpgradeProfile

if TYPE_CHECKING:
    import pathlib

    from uv_upx.services.collect_dependencies.models import DependencyGroupParsed
    from uv_upx.services.dependencies_from_project import DependenciesRegistry
    from uv_upx.services.dependency_up import ChangesList
    from uv_upx.services.dependency_up.models.changes_list import ChangesItem
    from uv_upx.services.parse_v2.collect_dependencies import CollectedTopLevelDependencies, PyProjectWrapperExtra


def ask_interactive_confirmation(
    changes_item: ChangesItem,
    group: DependencyGroupParsed,
    path: pathlib.Path,
) -> bool:
    """Ask the user for confirmation in interactive mode."""
    print("=" * 40)
    print(f"In file: {path.as_uri()}")
    folder_name = path.parent.name
    print_formatted_text(HTML(f"Parent folder: <b>{folder_name}</b>"))

    group_title = str(group.section)
    if group.group_name:
        group_title += f"[{group.group_name}]"
    print_formatted_text(HTML(f"Group: <b>{group_title}</b>"))

    message = HTML(
        f"<ansiblue><b>{escape(changes_item.from_item.get_name_with_extras())}:</b></ansiblue> "
        f"<ansired>{escape(changes_item.from_item.get_partial_spec())}</ansired>"
        f" <ansiyellow>&#8594;</ansiyellow> "
        f"<ansigreen>{escape(changes_item.to_item.get_partial_spec())}</ansigreen>",
    )

    result = choice(
        message=message,
        options=[
            (True, "Accept"),
            (False, "Reject"),
        ],
        default=True,
        #
        enable_interrupt=True,
    )

    print("=" * 40)

    return result


def handle_py_project_v2(
    *,
    dependencies_registry: DependenciesRegistry,
    py_project: PyProjectWrapperExtra,
    #
    verbose: bool,
    #
    profile: UpgradeProfile,
    #
    interactive: bool = False,
) -> ChangesList:
    """Handle a single pyproject.toml file."""
    logger = logging.getLogger(__name__)

    data = py_project.data

    changes: ChangesList = []

    # TODO: Optimize.

    for group in py_project.dependency_groups_parsed:
        for dependency in group.parsed_dependencies:
            dependency_parsed = dependency.parsed

            dependency_candidate = copy.deepcopy(dependency_parsed)

            change_or_none = update_dependency_v2(
                dependencies_registry=dependencies_registry,
                parsed=dependency_candidate,
                #
                profile=profile,
            )
            if change_or_none is not None:
                apply_change = True
                if interactive:
                    apply_change = ask_interactive_confirmation(
                        changes_item=change_or_none,
                        group=group,
                        path=py_project.path,
                    )

                if apply_change:
                    # TODO: Rework this to avoid double update.
                    change_or_none = update_dependency_v2(
                        dependencies_registry=dependencies_registry,
                        parsed=dependency_parsed,
                        #
                        profile=profile,
                    )
                    if change_or_none is None:
                        msg = "Change should be applied, but it is not detected."
                        raise RuntimeError(msg)

                    changes.append(change_or_none)

                    group.dependencies[dependency.index_in_group] = dependency_parsed.get_full_spec()

                elif verbose:
                    logger.info(
                        f"Skipped change for {dependency_parsed.package_name} in {py_project.path.as_uri()}",
                    )

    if changes or (profile is UpgradeProfile.WITH_PINNED):
        # Note: "(profile is UpgradeProfile.WITH_PINNED)" require some way to write back changes.
        toml_save(py_project.path, data)
        logger.info(f"Saved changes to {py_project.path.as_uri()}")

        for change in changes:
            logger.info(f"  {change}")

    return changes


def handle_py_projects_v2(
    *,
    dependencies_registry: DependenciesRegistry,
    collected_top_level_dependencies: CollectedTopLevelDependencies,
    #
    verbose: bool,
    #
    profile: UpgradeProfile,
    #
    interactive: bool = False,
) -> ChangesList:
    """Handle multiple pyproject.toml files."""
    changes: ChangesList = []

    if interactive:
        message = """You are running in interactive mode.
You will be prompted for each proposed change.

Note: These changes related to updating dependencies in pyproject.toml files.
It is implying how dependencies will be updated.
But, if you reject ">=X.Y.Z" change, it will not updated in pyproject.toml,
but it may still be updated in the lock file (e.g., uv.lock).
"""
        print(message)

    for py_project in collected_top_level_dependencies.parsed_pyprojects:
        changes_local = handle_py_project_v2(
            dependencies_registry=dependencies_registry,
            py_project=py_project,
            #
            verbose=verbose,
            #
            profile=profile,
            #
            interactive=interactive,
        )
        changes.extend(changes_local)

    return changes



import logging


def init_logging() -> None:
    logging.basicConfig(
        # Note: If needed "debug", more complex logging setup must be done.
        #   Because some libraries may be too verbose in debug mode.
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )



import pytest

from uv_upx.services.dependencies_from_project import Version
from uv_upx.services.dependency_up.models.dependency_parsed import VersionConstraint
from uv_upx.services.dependency_up.update_dependency import handle_version_constraint


@pytest.mark.parametrize(
    ("version_constraint", "version_new", "expected"),
    [
        (  # Main case. We focus mainly on this case.
            VersionConstraint(
                operator=">=",
                version="32.0",
            ),
            "33.1",
            ">=33.1",
        ),
        (  # Need to get a previous version if we want to implement an update.
            #   This includes handling alpha/beta/rc versions. So, just skip.
            VersionConstraint(
                operator=">",
                version="32.0",
            ),
            "33.1",
            ">32.0",
        ),
        (  # Not possible by uv. But check just in case.
            VersionConstraint(
                operator="==",
                version="32.0",
            ),
            "33.1",
            "==32.0",
        ),
        (  # Not possible by uv. But check just in case.
            VersionConstraint(
                operator="<",
                version="32.0",
            ),
            "33.1",
            "<32.0",
        ),
    ],
)
def test_handle_version_constraint(
    version_constraint: VersionConstraint,
    version_new: str,
    expected: str,
) -> None:
    handle_version_constraint(
        version_constraint=version_constraint,
        version_new=Version(version_new),
    )
    assert str(version_constraint) == expected



import copy
from typing import TYPE_CHECKING

from uv_upx.services.dependency_up.constants.operators import (
    VERSION_OPERATOR_I_GREATER_OR_EQUAL,
    VERSION_OPERATORS_I_PINNED_ALLOWED_TO_CHANGE,
)
from uv_upx.services.toml import toml_save

if TYPE_CHECKING:
    from uv_upx.services.parse_v2.collect_dependencies import CollectedTopLevelDependencies


def change_pinned_constraints(
    collected_top_level_dependencies: CollectedTopLevelDependencies,
    #
) -> None:
    copied_dependencies = copy.deepcopy(collected_top_level_dependencies)

    for py_project in copied_dependencies.parsed_pyprojects:
        is_have_changes_for_file = False

        for group in py_project.dependency_groups_parsed:
            for dependency in group.parsed_dependencies:
                dependency_parsed = dependency.parsed

                is_have_changes_for_dependency = False
                for version_constraint in dependency_parsed.version_constraints:
                    if version_constraint.operator in VERSION_OPERATORS_I_PINNED_ALLOWED_TO_CHANGE:
                        version_constraint.operator = VERSION_OPERATOR_I_GREATER_OR_EQUAL
                        is_have_changes_for_dependency = True

                if is_have_changes_for_dependency:
                    group.dependencies[dependency.index_in_group] = dependency_parsed.get_full_spec()
                    is_have_changes_for_file = True

        if is_have_changes_for_file:
            toml_save(py_project.path, py_project.data)



import subprocess
from typing import TYPE_CHECKING

from uv_upx.services.run_uv_related.exceptions import UnresolvedDependencyError

if TYPE_CHECKING:
    import pathlib


def run_uv_lock(
    workdir: pathlib.Path,
    *,
    upgrade: bool = False,
) -> None:
    # uv lock --upgrade
    command = ["uv", "lock"]
    if upgrade:
        command.append("--upgrade")
    try:
        subprocess.run(  # noqa: S603
            # uv lock --upgrade
            command,
            check=True,
            cwd=workdir,
        )
    except subprocess.CalledProcessError as e:
        msg = "Failed to resolve dependencies with 'uv lock'. Please check your dependency specifications."
        raise UnresolvedDependencyError(
            msg,
        ) from e



from uv_upx.services.dependencies_from_project.models import DependenciesRegistry, Version
from uv_upx.services.package_name import PackageName
from uv_upx.services.toml import toml_parse


def parse_from_uv_lock_file(
    content: str,
) -> DependenciesRegistry:
    data = toml_parse(content)

    dependencies = DependenciesRegistry()
    for package in data.get("package", []):  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
        version = package.get("version")  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

        if version is None:
            # Just in case. Possible problem with "[tool.hatch.version]"
            # https://github.com/zundertj/uv-bump/issues/5
            continue

        dependencies[PackageName(package["name"])] = Version(version)  # pyright: ignore[reportUnknownArgumentType] # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]

    return dependencies



import pathlib  # noqa: TC003
from typing import Annotated

import typer

from uv_upx.services.collect_top_level_dependencies.collect_top_level_dependencies import collect_top_level_dependencies
from uv_upx.services.normalize_paths import normalize_and_check_path_to_project_root

app = typer.Typer()


@app.command()
def collect_top_level_dependencies_from_project(
    *,
    project_root_path: Annotated[
        pathlib.Path | None,
        typer.Option(
            "--project",
            "-p",
            help="Path to project root directory. Use current working directory if not specified.",
        ),
    ] = None,
    #
    only_special_cases: Annotated[
        bool,
        typer.Option(
            "--only-special-cases",
            help="Collect only complex and unhandled dependencies",
        ),
    ] = False,
    preserve_original_package_names: Annotated[
        bool,
        typer.Option("--preserve-original-package-names", help="Preserve original package names in pyproject.toml"),
    ] = False,
) -> None:
    """Collect top-level dependencies from the project."""
    collect_top_level_dependencies(
        project_root_path=normalize_and_check_path_to_project_root(project_root_path),
        #
        only_special_cases=only_special_cases,
        #
        preserve_original_package_names=preserve_original_package_names,
    )



from .get_all_pyprojects import get_all_pyprojects_by_project_root_path
from .models import PyProjectsRegistry, PyProjectWrapper

__all__ = [
    "PyProjectWrapper",
    "PyProjectsRegistry",
    "get_all_pyprojects_by_project_root_path",
]



from pydantic import BaseModel

from uv_upx.services.dependency_up.models.dependency_parsed import DependencyParsed


class ChangesItem(BaseModel):
    from_item: DependencyParsed
    to_item: DependencyParsed

    def __str__(self) -> str:
        return (
            f"{self.from_item.get_name_with_extras()}: "
            f"{self.from_item.get_partial_spec()}"
            f" -> "
            f"{self.to_item.get_partial_spec()}"
        )


type ChangesList = list[ChangesItem]



import logging
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from uv_upx.services.collect_dependencies.collect_groups_from_py_project import collect_from_py_project
from uv_upx.services.collect_dependencies.models import (
    DependencyGroupParsed,
    DependencyItemParsed,
)
from uv_upx.services.dependency_up.parse_dependency import parse_dependency
from uv_upx.services.get_all_pyprojects import PyProjectWrapper, get_all_pyprojects_by_project_root_path

if TYPE_CHECKING:
    import pathlib


class PyProjectWrapperExtra(PyProjectWrapper):
    dependency_groups_parsed: list[DependencyGroupParsed] = Field(
        default_factory=list,
    )


class CollectedTopLevelDependencies(BaseModel):
    parsed_pyprojects: list[PyProjectWrapperExtra]


def collect_top_level_dependencies(
    *,
    project_root_path: pathlib.Path,
    #
    preserve_original_package_names: bool = False,
    #
    verbose: bool = False,
) -> CollectedTopLevelDependencies:
    """Collect top-level dependencies from all pyproject.toml files in the project."""
    logger = logging.getLogger(__name__)

    if verbose:
        logger.info(f"Collect dependencies from project at: {project_root_path.as_uri()}")

    py_projects = get_all_pyprojects_by_project_root_path(project_root_path)

    parsed_pyprojects: list[PyProjectWrapperExtra] = []
    for py_project in py_projects.items:
        dependency_groups_parsed: list[DependencyGroupParsed] = []

        for group in collect_from_py_project(py_project.data):
            # Note: With this we can show the string representation of dependencies. With comments.
            # print(group.dependencies.as_string())

            parsed_dependencies: list[DependencyItemParsed] = []

            for index, dependency in enumerate(group.dependencies):
                if not isinstance(dependency, str):
                    if verbose:
                        # https://docs.astral.sh/uv/concepts/projects/dependencies/#nesting-groups
                        logger.warning(f"Skipping non-string dependency: {dependency}")
                    continue

                parsed = parse_dependency(
                    dependency,
                    preserve_original_package_names=preserve_original_package_names,
                )

                parsed_dependencies.append(
                    DependencyItemParsed(
                        parsed=parsed,
                        index_in_group=index,
                    ),
                )

            if not parsed_dependencies:
                continue

            dependency_group_parsed = DependencyGroupParsed(
                section=group.section,
                group_name=group.group_name,
                dependencies=group.dependencies,
                #
                parsed_dependencies=parsed_dependencies,
            )
            dependency_groups_parsed.append(dependency_group_parsed)

        if not dependency_groups_parsed:
            continue

        py_project_extra = PyProjectWrapperExtra(
            path=py_project.path,
            data=py_project.data,
            #
            dependency_groups_parsed=dependency_groups_parsed,
        )
        parsed_pyprojects.append(py_project_extra)

    return CollectedTopLevelDependencies(
        parsed_pyprojects=parsed_pyprojects,
    )


# for index, _dependency in enumerate(deps_sequence_from_config):
#     if changes_or_none is not None:
#         deps_sequence_from_config[index] = changes_or_none.to_item.get_full_spec()



import pytest

from uv_upx.services.package_name import normalize_package_name


@pytest.mark.parametrize(
    ("package_name", "expected"),
    [
        ("foo", "foo"),
        ("Foo", "foo"),
        ("FoO", "foo"),
        ("FoO-bar", "foo-bar"),
        ("FoO_bar", "foo-bar"),
        #
        ("TOMLKit", "tomlkit"),
        ("Pydantic", "pydantic"),
        ("pyTEST_BenchMark", "pytest-benchmark"),
    ],
)
def test_package_name_normalization(
    package_name: str,
    expected: str,
) -> None:
    result = normalize_package_name(package_name)
    assert result == expected



import pytest


@pytest.fixture
def lock_file_contents() -> str:
    return """[[package]]
name = "bla"
version = "0.2.1"

[[package]]
name = "foo"
version = "1.21.0"
"""


@pytest.fixture
def pyproject_toml_contents() -> str:
    return """[project]
name = "uv-upx"
version = "0.2.4"

requires-python = ">=3.14"
dependencies = [
    # Better classes and data validation
    "pydantic>=2.12.5",
    # TOML parser and writer with preserved formatting
    "tomlkit>=0.13.3",
    # CLI app framework
    "typer>=0.20.0",
]
"""


# TODO: Add tests.



from typing import TYPE_CHECKING

from .parse_from_uv_lock import parse_from_uv_lock

if TYPE_CHECKING:
    import pathlib

    from . import DependenciesRegistry


def get_dependencies_from_project(
    workdir: pathlib.Path,
) -> DependenciesRegistry:
    """Get dependencies info.

    Concrete implementation can be changed.
    """
    return parse_from_uv_lock(workdir=workdir)



import logging
from typing import TYPE_CHECKING

from uv_upx.services.dependencies_from_project import get_dependencies_from_project
from uv_upx.services.dependency_up.handle_groups import handle_py_projects_v2
from uv_upx.services.get_all_pyprojects import get_all_pyprojects_by_project_root_path
from uv_upx.services.normalize_paths import get_and_check_path_to_uv_lock
from uv_upx.services.parse_v2.change_pinned_constraints import change_pinned_constraints
from uv_upx.services.parse_v2.collect_dependencies import collect_top_level_dependencies
from uv_upx.services.toml import toml_load
from uv_upx.services.updater.finalize_updating import finalize_updating
from uv_upx.services.updater.rollback_updater import RollbackData, rollback_updater
from uv_upx.services.updater.update_lock_file import update_lock_file
from uv_upx.services.upgrade_profile import UpgradeProfile

if TYPE_CHECKING:
    import pathlib


def run_updater(  # noqa: PLR0913
    *,
    project_root_path: pathlib.Path,
    #
    dry_run: bool = False,
    verbose: bool = False,
    #
    preserve_original_package_names: bool = False,
    #
    no_sync: bool = False,
    #
    interactive: bool = False,
    #
    profile: UpgradeProfile = UpgradeProfile.DEFAULT,
) -> None:
    """Orchestrates dependency updates with rollback on failure."""
    logger = logging.getLogger(__name__)

    uv_lock_path = get_and_check_path_to_uv_lock(project_root_path)
    uv_lock_data = toml_load(uv_lock_path)

    py_projects = get_all_pyprojects_by_project_root_path(project_root_path)
    if verbose:
        logger.info(f"Found {len(py_projects.items)} pyproject.toml files in the workspace.")
        for py_project in py_projects.items:
            logger.info(f"  {py_project.path.as_uri()}")

    rollback_data = RollbackData.from_parts(
        uv_lock_path=uv_lock_path,
        uv_lock_data=uv_lock_data,
        #
        py_projects=py_projects,
    )

    is_rollback_needed = dry_run
    rollback_message = "Rolling back to previous state because dry run is enabled."

    collected_top_level_dependencies = collect_top_level_dependencies(
        project_root_path=project_root_path,
        #
        preserve_original_package_names=preserve_original_package_names,
    )

    if profile is UpgradeProfile.WITH_PINNED:
        change_pinned_constraints(
            collected_top_level_dependencies=collected_top_level_dependencies,
        )

    try:
        update_lock_file(
            project_root_path,
        )

        dependencies_registry = get_dependencies_from_project(workdir=project_root_path)

        if handle_py_projects_v2(
            collected_top_level_dependencies=collected_top_level_dependencies,
            dependencies_registry=dependencies_registry,
            #
            verbose=verbose,
            #
            profile=profile,
            #
            interactive=interactive,
        ):
            logger.info("Updated pyproject.toml files successfully.")

            finalize_updating(
                project_root_path,
                dry_run=dry_run,
                #
                no_sync=no_sync,
                #
                profile=profile,
                #
                interactive=interactive,
            )

        else:
            msg = "No important changes detected. Rolling back to previous state."
            logger.info(msg)
            is_rollback_needed = True
            rollback_message = msg

    except Exception as e:  # noqa: BLE001
        msg = f"Failed to update dependencies: '{type(e)}:{e}' Rolling back to previous state."
        logger.error(msg)  # noqa: TRY400
        is_rollback_needed = True
        rollback_message = msg

    if dry_run:
        rollback_message = "Dry run enabled, rolling back to previous state."
        is_rollback_needed = True

    try:
        if is_rollback_needed:
            rollback_updater(
                rollback_data=rollback_data,
                #
                no_sync=no_sync,
            )
            logger.info(rollback_message)
    except Exception as e:  # noqa: BLE001
        msg = f"Failed to rollback: '{e}'"
        logger.error(msg)  # noqa: TRY400



import importlib
import pathlib  # noqa: TC003
from typing import Annotated

import typer

from uv_upx.services.normalize_paths import normalize_and_check_path_to_project_root
from uv_upx.services.updater import run_updater
from uv_upx.services.upgrade_profile import UpgradeProfile

app = typer.Typer(
    pretty_exceptions_enable=False,
    pretty_exceptions_show_locals=False,
)


def version_callback(
    value: bool,  # noqa: FBT001
) -> None:
    if value:
        app_name = "uv-upx"
        version: str = importlib.metadata.version(app_name)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType]
        typer.echo(f"{app_name} {version}")
        raise typer.Exit


# noinspection PyUnusedLocal
@app.command()
def run(  # noqa: PLR0913
    *,
    project_root_path: Annotated[
        pathlib.Path | None,
        typer.Option(
            "--project",
            "-p",
            help="Path to project root directory. Use current working directory if not specified.",
        ),
    ] = None,
    #
    dry_run: Annotated[bool, typer.Option("--dry-run", help="Show changes without writing file")] = False,
    #
    verbose: Annotated[bool, typer.Option("--verbose", help="Show more output")] = False,
    #
    preserve_original_package_names: Annotated[
        bool,
        typer.Option("--preserve-original-package-names", help="Preserve original package names in pyproject.toml"),
    ] = False,
    #
    no_sync: Annotated[
        bool,
        typer.Option(
            "--no-sync",
            help="Do not run uv-sync. "
            "In case of the complex build process. "
            "But, recommended to run with sync, for better chances for revealing problems.",
        ),
    ] = False,
    #
    profile: Annotated[
        UpgradeProfile | None,
        typer.Option(
            "--profile",
            help="Which profile to use when upgrading dependencies. (Experimental feature)",
        ),
    ] = None,
    #
    interactive: Annotated[
        bool,
        typer.Option(
            "--interactive",
            help="Enable interactive mode for selecting updates. (Experimental feature)",
        ),
    ] = False,
    #
    version: Annotated[  # noqa: ARG001  # pyright: ignore[reportUnusedParameter]
        bool | None,
        typer.Option(
            "--version",
            help="Show version and exit.",
            callback=version_callback,
            is_eager=True,
        ),
    ] = None,
) -> None:
    """Update pyproject.toml dependencies to latest compatible versions."""
    run_updater(
        project_root_path=normalize_and_check_path_to_project_root(project_root_path),
        #
        dry_run=dry_run,
        verbose=verbose,
        #
        preserve_original_package_names=preserve_original_package_names,
        #
        no_sync=no_sync,
        #
        interactive=interactive,
        #
        profile=profile or UpgradeProfile.get_default(),
    )



import pathlib

from pydantic import BaseModel, ConfigDict
from tomlkit import TOMLDocument

type PathToPyprojectToml = pathlib.Path
"""Path to a pyproject.toml file."""


class PyProjectWrapper(BaseModel):
    path: PathToPyprojectToml
    data: TOMLDocument

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class PyProjectsRegistry(BaseModel):
    items: list[PyProjectWrapper]



import copy
import logging
from typing import TYPE_CHECKING, Any

from uv_upx.services.dependency_up.constants.operators import (
    VERSION_OPERATOR_I_GREATER_OR_EQUAL,
    VERSION_OPERATORS_I_EXPLICIT_IGNORE,
    VERSION_OPERATORS_I_PINNED_ALLOWED_TO_CHANGE,
    VERSION_OPERATORS_I_PUT_IF_DIFFERENT,
)
from uv_upx.services.dependency_up.models.changes_list import ChangesItem
from uv_upx.services.dependency_up.models.dependency_parsed import DependencyParsed, VersionConstraint
from uv_upx.services.upgrade_profile import UpgradeProfile

if TYPE_CHECKING:
    from uv_upx.services.dependencies_from_project import DependenciesRegistry, Version


def update_dependency_v2(
    *,
    dependencies_registry: DependenciesRegistry,
    parsed: DependencyParsed,
    #
    profile: UpgradeProfile,
) -> ChangesItem | None:
    """Updates the dependency version based on registry lookup."""
    try:
        version_new = dependencies_registry[parsed.package_name]
    except KeyError:
        # Note: raise error, because it now we have all the dependencies in the registry.
        msg = f"Dependency not found in the registry: {parsed.package_name}"
        logger = logging.getLogger(__name__)
        logger.error(msg)  # noqa: TRY400
        return None

    parsed_original = copy.deepcopy(parsed)

    is_has_changes = False
    for version_constraint in parsed.version_constraints:
        is_has_changes_local = handle_version_constraint(
            version_constraint=version_constraint,
            version_new=version_new,
            #
            profile=profile,
        )
        if is_has_changes_local:
            is_has_changes = True

    if not parsed.version_constraints:
        parsed.version_constraints.append(
            VersionConstraint(operator=VERSION_OPERATOR_I_GREATER_OR_EQUAL, version=version_new),
        )
        is_has_changes = True

    if is_has_changes:
        return ChangesItem(
            from_item=parsed_original,
            to_item=parsed,
        )
    return None


type IsHasChanges = bool


def handle_version_constraint(
    *,
    version_constraint: VersionConstraint,
    version_new: Version,
    #
    profile: UpgradeProfile = UpgradeProfile.DEFAULT,
    #
    verbose: bool = False,
    dependency: Any | None = None,  # noqa: ANN401
) -> IsHasChanges:
    """Handle a single version constraint.

    Change it in-place. Because, for now, it is simpler than rebuilding the whole object.
    """
    logger = logging.getLogger(__name__)

    is_has_changes = False
    if version_constraint.operator in VERSION_OPERATORS_I_PUT_IF_DIFFERENT:
        # TODO: (?) Implement better version comparison logic here
        if version_constraint.version != version_new:
            version_constraint.version = version_new
            is_has_changes = True

    elif (profile is UpgradeProfile.WITH_PINNED) and (
        version_constraint.operator in VERSION_OPERATORS_I_PINNED_ALLOWED_TO_CHANGE
    ):
        # sourcery skip: hoist-similar-statement-from-if, hoist-statement-from-if
        if version_constraint.version != version_new:
            version_constraint.version = version_new
            is_has_changes = True
        # else:
        #     # Note: Workaround. Because we need to roll back the operator change.
        #     is_has_changes = True

    elif version_constraint.operator in VERSION_OPERATORS_I_EXPLICIT_IGNORE:
        if verbose and dependency:
            msg = f"Explicitly ignoring version constraint {version_constraint}. Dependency: {dependency}"
            logger.info(msg)
    elif verbose and dependency:
        msg = f"Operator {version_constraint.operator} is not supported yet. Skip. Dependency: {dependency}"
        logger.warning(msg)

    return is_has_changes



from uv_upx.cli.main import app as app_full
from uv_upx.cli.upgrade.main import app as app_upgrade
from uv_upx.logging_custom import init_logging


def main() -> None:
    init_logging()
    app_full()


def main_short() -> None:
    init_logging()
    app_upgrade()



from typing import TYPE_CHECKING

from uv_upx.services.run_uv_related import run_uv_lock

if TYPE_CHECKING:
    import pathlib


def update_lock_file(
    project_root_path: pathlib.Path,
) -> None:
    # Because we want a fast update. Without triggering build for now.
    run_uv_lock(
        workdir=project_root_path,
        upgrade=True,
    )



import copy
import logging
import pathlib

from pydantic import BaseModel, ConfigDict
from tomlkit import TOMLDocument

from uv_upx.services.get_all_pyprojects import PyProjectsRegistry
from uv_upx.services.run_uv_related import UvSyncMode, run_uv_sync
from uv_upx.services.toml import toml_save


class UvLockWrapper(BaseModel):
    path: pathlib.Path
    data: TOMLDocument

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class RollbackData(BaseModel):
    uv_lock: UvLockWrapper

    py_projects: PyProjectsRegistry

    @classmethod
    def from_parts(
        cls,
        *,
        uv_lock_path: pathlib.Path,
        uv_lock_data: TOMLDocument,
        #
        py_projects: PyProjectsRegistry,
    ) -> RollbackData:
        uv_lock_wrapper = UvLockWrapper(
            path=uv_lock_path,
            data=copy.deepcopy(uv_lock_data),
        )
        return cls(
            uv_lock=uv_lock_wrapper,
            py_projects=copy.deepcopy(py_projects),
        )


def rollback_updater(
    *,
    rollback_data: RollbackData,
    #
    no_sync: bool = False,
) -> None:
    logger = logging.getLogger(__name__)

    toml_save(rollback_data.uv_lock.path, rollback_data.uv_lock.data)
    for py_project in rollback_data.py_projects.items:
        toml_save(py_project.path, py_project.data)

    if not no_sync:
        run_uv_sync(
            workdir=rollback_data.uv_lock.path.parent,
            uv_sync_mode=UvSyncMode.FROZEN,
        )

    logger.info("Rollback completed.")



from tomlkit.items import Array

from uv_upx.services.dependency_up import IncludedDependencyGroup
from uv_upx.services.dependency_up.models.dependency_parsed import DependencyString

type TomlBasedDependenciesList = Array | list[DependencyString] | list[DependencyString | IncludedDependencyGroup]
"""List of dependencies from TOML document

Data needed to be changed directly in this list to preserve comments.

This is a reference to the part of the source TOML document.
"""



import re
from re import Pattern
from typing import Annotated, Final

from pydantic import AfterValidator, ConfigDict, RootModel

PATTERN_I_NORMALIZED_I_TO_REPLACE: Final[Pattern[str]] = re.compile(r"[-_.]+")
CHAR_I_TO_USE: Final[str] = "-"


def normalize_package_name(package_name: str) -> str:
    """Normalize the package name according to PEP 503.

    https://peps.python.org/pep-0503/

    https://packaging.python.org/en/latest/specifications/name-normalization/
    """
    return PATTERN_I_NORMALIZED_I_TO_REPLACE.sub(CHAR_I_TO_USE, package_name).lower()


class PackageName(RootModel[str]):
    root: Annotated[str, AfterValidator(normalize_package_name)]

    def __str__(self) -> str:
        return self.root

    model_config = ConfigDict(
        frozen=True,
    )



import logging
from typing import TYPE_CHECKING

from uv_upx.services.normalize_paths import get_and_check_path_to_uv_lock
from uv_upx.services.run_uv_related import UvSyncMode, run_uv_lock, run_uv_sync
from uv_upx.services.upgrade_profile import UpgradeProfile

if TYPE_CHECKING:
    import pathlib


def finalize_updating(
    project_root_path: pathlib.Path,
    *,
    dry_run: bool = False,
    #
    no_sync: bool = False,
    #
    profile: UpgradeProfile = UpgradeProfile.DEFAULT,
    #
    interactive: bool = False,
) -> None:
    logger = logging.getLogger(__name__)

    if dry_run:
        logger.info("Dry run. No changes were made.")
        return

    if profile == UpgradeProfile.WITH_PINNED or interactive:
        uv_lock_path = get_and_check_path_to_uv_lock(project_root_path)
        uv_lock_path.unlink(missing_ok=True)

    if no_sync:
        run_uv_lock(
            workdir=project_root_path,
        )

        logger.info("Updated uv.lock successfully.")
    else:
        # Because we want to re-check that all is ok.
        run_uv_sync(
            workdir=project_root_path,
            uv_sync_mode=UvSyncMode.DEFAULT,
        )
        logger.info("Synced dependencies successfully with updating uv.lock.")



import enum


@enum.unique
class UpgradeProfile(enum.StrEnum):
    DEFAULT = "default"
    """Default profile"""

    WITH_PINNED = "with_pinned"
    """Upgrade also "pinned" (== exact version) dependencies."""

    @staticmethod
    def get_default() -> UpgradeProfile:
        return UpgradeProfile.DEFAULT



from typing import TYPE_CHECKING

from uv_upx.services.dependencies_from_project.parse_from_uv_lock_file import parse_from_uv_lock_file
from uv_upx.services.normalize_paths import get_and_check_path_to_uv_lock

if TYPE_CHECKING:
    import pathlib

    from uv_upx.services.dependencies_from_project.models import DependenciesRegistry


def parse_from_uv_lock(
    workdir: pathlib.Path,
) -> DependenciesRegistry:
    return parse_from_uv_lock_file(
        content=get_and_check_path_to_uv_lock(
            workdir,
        ).read_text(),
    )



from .get_dependencies_from_project import get_dependencies_from_project
from .models import DependenciesRegistry, Version

__all__ = [
    "DependenciesRegistry",
    "Version",
    "get_dependencies_from_project",
]



import typer

from uv_upx.cli.helpers.main import app as app_helpers
from uv_upx.cli.upgrade.main import app as app_upgrade

app = typer.Typer(
    pretty_exceptions_enable=False,
    pretty_exceptions_show_locals=False,
)

app.add_typer(app_upgrade, name="upgrade")
app.add_typer(app_helpers, name="helpers")



from typing import NewType

from pydantic import Field, RootModel

from uv_upx.services.package_name import PackageName

Version = NewType("Version", str)

type DependenciesRegistryInternal = dict[PackageName, Version]


class DependenciesRegistry(RootModel[DependenciesRegistryInternal]):
    """Registry of dependencies with their versions.

    Source of truth for the dependencies.
    """

    root: DependenciesRegistryInternal = Field(default_factory=dict)

    def __getitem__(self, item: PackageName) -> Version:
        return self.root[item]

    def __setitem__(self, key: PackageName, value: Version) -> None:
        self.root[key] = value



from uv_upx.services.run_uv_related.exceptions import UnresolvedDependencyError
from uv_upx.services.run_uv_related.run_uv_lock import run_uv_lock
from uv_upx.services.run_uv_related.run_uv_sync import UvSyncMode, run_uv_sync

__all__ = [
    "UnresolvedDependencyError",
    "UvSyncMode",
    "run_uv_lock",
    "run_uv_sync",
]



import pytest

from uv_upx.services.dependency_up.models.dependency_parsed import DependencyParsed, VersionConstraint
from uv_upx.services.dependency_up.parse_dependency import parse_dependency
from uv_upx.services.package_name import PackageName


@pytest.mark.parametrize(
    ("dependency_string", "expected"),
    [
        (  # Only name
            "foo",
            DependencyParsed(
                package_name=PackageName("foo"),
            ),
        ),
        #
        (  # Simple version constraint
            "foo>=32.0",
            DependencyParsed(
                package_name=PackageName("foo"),
                version_constraints=[
                    VersionConstraint(
                        operator=">=",
                        version="32.0",
                    ),
                ],
            ),
        ),
        (  # Simple version constraint
            "foo<=32.0",
            DependencyParsed(
                package_name=PackageName("foo"),
                version_constraints=[
                    VersionConstraint(
                        operator="<=",
                        version="32.0",
                    ),
                ],
            ),
        ),
        (  # Multiple version constraints
            "foo[bla,xyz]>=32.0,!=33,<34;python_version<'3.11'",
            DependencyParsed(
                package_name=PackageName("foo"),
                version_constraints=[
                    VersionConstraint(
                        operator=">=",
                        version="32.0",
                    ),
                    VersionConstraint(
                        operator="!=",
                        version="33",
                    ),
                    VersionConstraint(
                        operator="<",
                        version="34",
                    ),
                ],
                extras=["bla", "xyz"],
                marker="python_version<'3.11'",
            ),
        ),
        (  # Multiple version constraints
            'foo[bla,xyz] >=32.0, !=33, <34 ;python_version<"3.11"',
            DependencyParsed(
                package_name=PackageName("foo"),
                version_constraints=[
                    VersionConstraint(
                        operator=">=",
                        version="32.0",
                    ),
                    VersionConstraint(
                        operator="!=",
                        version="33",
                    ),
                    VersionConstraint(
                        operator="<",
                        version="34",
                    ),
                ],
                extras=["bla", "xyz"],
                marker='python_version<"3.11"',
            ),
        ),
    ],
)
def test_parse_dependency(
    dependency_string: str,
    expected: DependencyParsed,
) -> None:
    result = parse_dependency(dependency_string)
    assert result == expected



import enum
import subprocess
from typing import TYPE_CHECKING

from uv_upx.services.run_uv_related import UnresolvedDependencyError

if TYPE_CHECKING:
    import pathlib


class UvSyncMode(enum.Enum):
    UPGRADE = enum.auto()
    FROZEN = enum.auto()
    DEFAULT = enum.auto()


def run_uv_sync(
    workdir: pathlib.Path,
    uv_sync_mode: UvSyncMode,
    *,
    include_all: bool = True,
) -> None:
    # uv sync --all-groups --all-extras --all-packages --frozen
    command = ["uv", "sync"]
    if include_all:
        command.extend(["--all-groups", "--all-extras", "--all-packages"])

    match uv_sync_mode:
        case UvSyncMode.UPGRADE:
            command.append("--upgrade")
        case UvSyncMode.FROZEN:
            command.append("--frozen")
        case UvSyncMode.DEFAULT:
            pass

    try:
        subprocess.run(  # noqa: S603
            command,
            check=True,
            cwd=workdir,
        )
    except subprocess.CalledProcessError as e:
        msg = "Failed to sync dependencies with 'uv sync'. Please check your dependency specifications."
        raise UnresolvedDependencyError(
            msg,
        ) from e



import itertools
import logging
from typing import TYPE_CHECKING, Any

from tomlkit.items import Array, Table

from uv_upx.services.collect_dependencies.models import DependencyGroup, DependencySection

if TYPE_CHECKING:
    from collections.abc import Iterable

    from tomlkit import TOMLDocument


# https://docs.astral.sh/uv/concepts/projects/dependencies/


def collect_i_main_dependency_group(
    data: TOMLDocument,
) -> Iterable[DependencyGroup]:
    """Check the main dependencies group."""
    logger = logging.getLogger(__name__)

    project: dict[Any, Any] = data.get("project", {})  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

    if not isinstance(project, dict):
        logger.warning("No [project] table found in pyproject.toml")
        return

    dependencies = project.get("dependencies", [])  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
    if not dependencies:
        return

    if not isinstance(dependencies, Array):
        logger.warning("No [project].dependencies found in pyproject.toml")
        return

    yield DependencyGroup(
        section=DependencySection.MAIN,
        dependencies=dependencies,  # pyright: ignore[reportUnknownArgumentType]
    )


def collect_i_dependency_groups(
    data: TOMLDocument,
) -> Iterable[DependencyGroup]:
    """Check the dependency-groups table.

    https://docs.astral.sh/uv/concepts/projects/dependencies/#development-dependencies
    """
    if groups := data.get("dependency-groups", {}):  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
        yield from collect_dependency_groups_i_many(
            groups=groups,  # pyright: ignore[reportUnknownArgumentType]
            section=DependencySection.DEPENDENCY_GROUPS,
        )
    else:
        return


def collect_i_optional_dependencies(
    data: TOMLDocument,
) -> Iterable[DependencyGroup]:
    """Check the project.optional-dependencies table.

    https://docs.astral.sh/uv/concepts/projects/dependencies/#optional-dependencies
    """
    if groups := data.get("project", {}).get("optional-dependencies", {}):  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
        yield from collect_dependency_groups_i_many(
            groups=groups,  # pyright: ignore[reportUnknownArgumentType]
            section=DependencySection.OPTIONAL_DEPENDENCIES,
        )
    else:
        return


def collect_dependency_groups_i_many(
    *,
    groups: Table,
    #
    section: DependencySection,
) -> Iterable[DependencyGroup]:
    """Check the dependency-groups table."""
    for group_name, _group_val in groups.items():  # pyright: ignore[reportUnknownVariableType]
        if not isinstance(_group_val, Array):
            continue

        group_val = _group_val  # pyright: ignore[reportUnknownVariableType]

        yield DependencyGroup(
            section=section,
            group_name=group_name,  # pyright: ignore[reportUnknownArgumentType]
            dependencies=group_val,
        )


def collect_from_py_project(
    data: TOMLDocument,
) -> Iterable[DependencyGroup]:
    yield from itertools.chain(
        collect_i_main_dependency_group(data=data),
        collect_i_dependency_groups(data=data),
        collect_i_optional_dependencies(data=data),
    )



from typing import Final

# https://peps.python.org/pep-0440/#version-specifiers

type VERSION_OPERATOR = str

VERSION_OPERATOR_I_GREATER_OR_EQUAL: Final[VERSION_OPERATOR] = ">="

VERSION_OPERATOR_I_EQUAL: Final[VERSION_OPERATOR] = "=="

VERSION_OPERATORS_I_PUT_IF_DIFFERENT: Final[set[VERSION_OPERATOR]] = {
    VERSION_OPERATOR_I_GREATER_OR_EQUAL,
}

VERSION_OPERATORS_I_PINNED_ALLOWED_TO_CHANGE: Final[set[VERSION_OPERATOR]] = {
    VERSION_OPERATOR_I_EQUAL,
}

VERSION_OPERATORS_I_EXPLICIT_IGNORE: Final[set[VERSION_OPERATOR]] = {
    # Pinned versions
    VERSION_OPERATOR_I_EQUAL,
    "===",
    #
    # Upper bounds
    "<",
    "<=",
    #
    # May needed advanced logic here
    "~=",
    #
    # Need to calculate a previous version. Skip for now.
    ">",
    #
    # Special case
    "!=",
}

VERSION_OPERATORS_I_ALL: Final[set[VERSION_OPERATOR]] = (
    VERSION_OPERATORS_I_PUT_IF_DIFFERENT | VERSION_OPERATORS_I_EXPLICIT_IGNORE
)



import re
from re import Pattern
from typing import Final

from uv_upx.services.dependency_up.constants.operators import VERSION_OPERATORS_I_ALL
from uv_upx.services.dependency_up.models.dependency_parsed import DependencyParsed, DependencyString, VersionConstraint
from uv_upx.services.package_name import PackageName

VERSION_OPERATORS_AS_OR: Final[str] = "|".join(sorted(VERSION_OPERATORS_I_ALL, key=lambda x: len(x), reverse=True))

# https://peps.python.org/pep-0440/#version-specifiers
PATTERN_I_DEPENDENCY_STRING: Final[Pattern[str]] = re.compile(
    rf"""^
\s*
(?P<name>[A-Za-z0-9_.+-]+)                       # package name (letters, digits, _ . + -)
\s*
(?:\[(?P<extras>[^\]]+)\])?                       # optional extras inside [...]
\s*
(?P<version_constraints>                           # optional version constraints (comma separated)
    (?:
        (?:(?:{VERSION_OPERATORS_AS_OR})\s*[^,;\[]+)    # single constraint starting with an operator
        (?:\s*,\s*(?:(?:{VERSION_OPERATORS_AS_OR})\s*[^,;\[]+))*  # optional additional constraints
    )
)?
\s*
(?:;\s*(?P<marker>.*))?                           # optional environment marker after ';'
$
""",
    re.IGNORECASE | re.VERBOSE,
)

PATTERN_I_VERSION_CONSTRAINT: Final[Pattern[str]] = re.compile(
    rf"""^
\s*
(?P<operator>{VERSION_OPERATORS_AS_OR})          # version operator
\s*
(?P<version>[^{VERSION_OPERATORS_AS_OR}]+)                # version value
\s*$
""",
    re.IGNORECASE | re.VERBOSE,
)


def parse_dependency(
    dependency_string: DependencyString,
    #
    *,
    preserve_original_package_names: bool = False,
) -> DependencyParsed:  # sourcery skip: low-code-quality
    dependency_string = dependency_string.strip()

    match = PATTERN_I_DEPENDENCY_STRING.match(dependency_string)

    if not match:
        msg = f"Invalid dependency string: {dependency_string}"
        raise ValueError(msg)

    name = match.group("name")

    extras_raw = match.group("extras") or ""
    extras = [item_ for item in extras_raw.split(",") if (item_ := item.strip())]

    version_constraints_raw = match.group("version_constraints") or ""
    version_constraints = parse_version_constraints(version_constraints_raw)

    marker_raw = match.group("marker")
    marker = marker_raw.strip() if marker_raw else None

    return DependencyParsed(
        original_name=name if preserve_original_package_names else None,
        package_name=PackageName(name),
        extras=extras,
        version_constraints=version_constraints,
        marker=marker,
    )


def parse_version_constraints(
    version_part: str,
) -> list[VersionConstraint]:
    # parse version constraints (comma separated)
    version_constraints: list[VersionConstraint] = []

    if not version_part:
        return version_constraints

    # do not attempt to parse direct URL / VCS refs (start with '@')
    if version_part.startswith("@"):
        msg = f"Invalid version_part string: '{version_part}'"
        raise ValueError(msg)

    raw_parts = [p.strip() for p in version_part.split(",") if p.strip()]

    for raw_part in raw_parts:
        match_vc = PATTERN_I_VERSION_CONSTRAINT.match(raw_part.strip())
        if not match_vc:
            msg = f"Invalid version constraint: '{raw_part}'"
            raise ValueError(msg)

        operator = match_vc.group("operator")
        version = match_vc.group("version")

        if operator not in VERSION_OPERATORS_I_ALL:
            msg = f"Invalid version operator: '{operator}'"
            raise ValueError(msg)

        version_constraint = VersionConstraint(
            operator=operator,
            version=version,
        )
        version_constraints.append(version_constraint)

    return version_constraints



import enum

from pydantic import BaseModel, ConfigDict

from uv_upx.services.dependency_up.models.dependencies_list import TomlBasedDependenciesList
from uv_upx.services.dependency_up.models.dependency_parsed import DependencyParsed


class DependencySection(enum.StrEnum):
    MAIN = "project.dependencies"
    DEPENDENCY_GROUPS = "dependency-groups"
    OPTIONAL_DEPENDENCIES = "project.optional-dependencies"


type GroupNameOrNone = str | None


class DependencyGroup(BaseModel):
    section: DependencySection
    dependencies: TomlBasedDependenciesList

    group_name: GroupNameOrNone = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class DependencyItemParsed(BaseModel):
    parsed: DependencyParsed
    index_in_group: int


class DependencyGroupParsed(DependencyGroup):
    parsed_dependencies: list[DependencyItemParsed]



from uv_upx.services.dependency_up.models.changes_list import ChangesList
from uv_upx.services.dependency_up.models.included_dependency_group import IncludedDependencyGroup

__all__ = [
    "ChangesList",
    "IncludedDependencyGroup",
]



from typing import Annotated

from pydantic import AfterValidator, BaseModel, Field

from uv_upx.services.dependency_up.constants.operators import VERSION_OPERATOR, VERSION_OPERATORS_I_ALL
from uv_upx.services.package_name import PackageName

# TODO: (?) Use this?
#   from packaging.specifiers import SpecifierSet
#   from packaging.version import Version


def validate_version_value(value: str) -> str:
    if not value.strip():
        msg = "Version value cannot be empty or whitespace"
        raise ValueError(msg)

    if not value[0].isdigit():
        msg = "Version value must start with a digit"
        raise ValueError(msg)

    return value


def validate_operator(value: str) -> VERSION_OPERATOR:
    if value not in VERSION_OPERATORS_I_ALL:
        msg = f"Operator must be one of {VERSION_OPERATORS_I_ALL}"
        raise ValueError(msg)
    return value


class VersionConstraint(BaseModel):
    operator: Annotated[VERSION_OPERATOR, AfterValidator(validate_operator)]
    version: Annotated[str, AfterValidator(validate_version_value)]

    def __str__(self) -> str:
        return f"{self.operator}{self.version}"


type DependencyString = str
"""Dependency string from pyproject.toml

Like:
- requests[dev] >=1.2.3; python_version >= "3.10"
- faker>=30.1
"""


class DependencyParsed(BaseModel):
    # https://peps.python.org/pep-0508/

    original_name: str | None = None
    """Original dependency name (e.g., reQuests).

    None if not preserved.
    """

    package_name: PackageName
    """Normalized package name (e.g., requests).

    Needed for better search.
    """

    extras: list[str] = Field(default_factory=list)
    """Extras (e.g., [dev])"""

    version_constraints: list[VersionConstraint] = Field(default_factory=list)
    """Version constraints (e.g., `>=1.2.3`, `==4.5.6`)"""

    marker: str | None = None
    """Environment marker (after ;)"""

    def get_name(self) -> str:
        if self.original_name is not None:
            return self.original_name
        return str(self.package_name)

    def get_name_with_extras(
        self,
    ) -> str:
        parts: list[str] = [self.get_name()]
        if self.extras:
            extras_str = ",".join(self.extras)
            parts.append(f"[{extras_str}]")
        return "".join(parts)

    def get_full_spec(
        self,
    ) -> str:
        parts: list[str] = [
            self.get_name_with_extras(),
            self.get_partial_spec(),
        ]

        return "".join(parts)

    def get_partial_spec(
        self,
    ) -> str:
        parts: list[str] = []
        if self.version_constraints:
            vc_strs = [f"{vc!s}" for vc in self.version_constraints]
            parts.append(",".join(vc_strs))
        if self.marker:
            parts.append(f"; {self.marker}")
        return "".join(parts)
